(* $Id: gontmodule.ml,v 1.3 2001/11/18 23:15:43 malekith Exp $ *)

open Gontutil

type module_symbol = {ms_name: id;
                      ms_type: ty;
                      mutable ms_symbol: symbol option}
                      
type mod_record = {mod_name: id;
                   mod_decls: module_symbol list}

let ver = "Gont Compiled Interface, $Revision: 1.3 $"

let save m out_ch decls =
  let make_module_symbol decl =
    match decl with
      GlobalDecl {global_type = t; global_name = n}
    | FuncProto {fp_type = t; fp_name = n} ->
        {ms_name = n; ms_type = t; ms_symbol = None}
    | FuncDecl _
    | ExternC _
    | TopLevelOpen _ -> assert false in
  let r = {mod_name = m; mod_decls = (List.map make_module_symbol decls)} in
  Marshal.to_channel out_ch ver [];
  Marshal.to_channel out_ch r []
;;

let modules : (string, mod_record) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear modules

exception Cannot_open
    
let open_module m =
  let load m =
    let rec open_any names =
      match names with
          x::xs -> 
            (try
              open_in x
             with 
               Sys_error _ -> open_any xs)
        | [] -> raise Cannot_open in
    let in_ch = 
      try
        open_any [((String.uncapitalize m) ^ ".gio")]
      with Cannot_open -> 
        raise (Error (xf "cannot find module interface for %s" m)) in
    let id = (Marshal.from_channel in_ch : string) in
      if id <> ver then
        (close_in in_ch; 
         raise (Error (xf "module interface file for %s was compiled \
                           with diffrent version, please recompile \
                           the interface" m)))
      else
        let mr = (Marshal.from_channel in_ch : mod_record) in
          if mr.mod_name <> m then
            (close_in in_ch;
             raise (Error (xf "interface file contains module %s, \
                               instead of expected %s" mr.mod_name m)))
           else
             Hashtbl.add modules mr.mod_name mr;
             close_in in_ch in
  try
    ignore(Hashtbl.find modules m)
  with Not_found -> load m
;;

let find_module m =
  open_module m;
  Hashtbl.find modules m

(* first string is module name, second -- symbol name *)
exception Has_no_symbol of ty

let import_symbol mod_name sym_name =
  let m = find_module mod_name in
  let rec find l =
    match l with
        {ms_name = n; ms_symbol = Some sym}::_ when (n = sym_name) -> 
          sym
      | {ms_name = n; ms_type = t}::_ when (n = sym_name) -> 
          raise (Has_no_symbol t)
      | _::xs -> 
          find xs
      | [] -> 
          raise Not_found in
   find m.mod_decls
  
let assign_symbol mod_name sym_name sym =
  let m = find_module mod_name in
  let rec find l =
    match l with
        {ms_name = n; ms_symbol = Some _}::_ when (n = sym_name) -> 
          assert false
      | ({ms_name = n} as r)::_ when (n = sym_name) -> 
          r.ms_symbol <- Some sym
      | _::xs -> 
          find xs
      | [] -> 
          raise Not_found in
   find m.mod_decls

let globals mod_name =
  let m = find_module mod_name in
  List.map (fun {ms_name = n; ms_type = t} -> (t, n)) m.mod_decls
;;
